import os
import time
import random
import asyncio
import aiohttp
import json
import traceback

from .multi_process_mng import MultiProcessMng, ProducerProc, ConsumerProc, defautl_logger_setup, ParallelConfig
from .aio_task_runner import AioTaskRunner

class AioParallelConfig(ParallelConfig):
    def __init__(self, producer_num=5, consumer_num=2, aio_parallel_num=5, one_parallel_task_num=-1, 
                 queue_num=1, queue_size=100, no_wait=False, timeout=300, run_time_len=-1, max_task_num=-1):
        super().__init__(producer_num, consumer_num, queue_num, queue_size, run_time_len, max_task_num)
        self.aio_parallel_num = aio_parallel_num
        self.one_parallel_task_num = one_parallel_task_num
        self.no_wait = no_wait
        self.timeout = timeout
    def add_argument(self, parser):
        super().add_argument(parser)
        parser.add_argument('--aio_parallel_num', type=int, default=5, help="")
        parser.add_argument('--one_parallel_task_num', type=int, default=-1, help="")
        parser.add_argument('--no_wait', type=str, default='false', help="")
        parser.add_argument('--timeout', type=int, default=300, help="")
    def fromParser(self, args):
        super().from_parser(args)
        self.aio_parallel_num = args.aio_parallel_num
        self.one_parallel_task_num = args.one_parallel_task_num
        self.no_wait = args.no_wait.lower() == 'true'
        self.timeout = args.timeout

class ReqDataBuildProc(ProducerProc):
    def pre_start(self):
        self.host = self.user_data['data'].get('host', 'http://127.0.0.1:8080')
    def step_run(self):
        url = f'{self.host}/search'
        header = {'Content-Type': 'application/json'}
        body = json.dumps({'test': 123})
        time.sleep(random.random())
        self.queue.put([url, header, body])
        return 'success', False

class AioHttpRunProc(ConsumerProc):
    def pre_start(self):
        self.parallel_cfg:AioParallelConfig = self.user_data['_sys_cfg']['parallel_cfg']
        # self.runner = AioTaskRunner(self.parallel_cfg.one_parallel_task_num, lambda para: self.put_counter(force=True))
        self.runner = AioTaskRunner(self.parallel_cfg.one_parallel_task_num)
    def run(self):
        try:
            if self.parallel_cfg.no_wait:
                asyncio.run(self.runner.run_times_per_second(self.parallel_cfg.aio_parallel_num, self._task_proc))
            else:
                asyncio.run(self.runner.parallel_run(self.parallel_cfg.aio_parallel_num, self._task_proc))
        except KeyboardInterrupt:
            self.runner.stop()
        except asyncio.CancelledError:
            pass
    async def _task_proc(self, session: aiohttp.ClientSession, parallet_id, task_idx, user_data):
        if self.stop_event.is_set():
            self.runner.stop_flag = True
            return
        req_data = self.get_data()
        if req_data is None:
            await asyncio.sleep(0.01)
            return
        req_start_time = time.time()
        self.counter_info.rec_start()
        rslt = await self._req(session, req_data)
        self.counter_info.rec_end(rslt, time.time() - req_start_time)
    async def _req(self, session: aiohttp.ClientSession, req_data):
        try:
            rsp = await session.post(url=req_data[0], headers=req_data[1], data=req_data[2], timeout=self.parallel_cfg.timeout, ssl=False)
            data = json.loads(await rsp.text())
            if rsp.status != 200 or data['code'] != 'LMS.00000000':
                return 'fail'
        except Exception as e:
            self.error(f'excet: {traceback.format_exc()}')
            await asyncio.sleep(0.01)
            return 'except'
        return 'success'

def run_aio_parallel_task(parallel_cfg:AioParallelConfig, ProducerProcClass=ReqDataBuildProc, ConsumerProcClass=AioHttpRunProc,
    data={}, callback=None, callback_interval=1, msg_callback=None, log_counter=True):
    _sys_cfg = {'parallel_cfg': parallel_cfg}
    mng = MultiProcessMng(parallel_cfg.queue_size, parallel_cfg.queue_num, log_counter=log_counter)
    try:
        mng.start_producer(ProducerProcClass, parallel_cfg.producer_num, {'_sys_cfg': _sys_cfg, 'data': data})
        time.sleep(3)
        mng.start_consumer(ConsumerProcClass, parallel_cfg.consumer_num, {'_sys_cfg': _sys_cfg, 'data': data})
        time.sleep(2)
        mng.run(max_task_num=parallel_cfg.max_task_num, run_time_len=parallel_cfg.run_time_len, 
            callback=callback, callback_interval=callback_interval, msg_callback=msg_callback)
    except KeyboardInterrupt:
        pass
    finally:
        mng.stop()
    return mng

class DemoAioHttpRunProc(AioHttpRunProc):
    async def _req(self, session: aiohttp.ClientSession, req_data):
        # print(req_data)
        try:
            await asyncio.sleep(random.random() * 2)
        except:
            await asyncio.sleep(0.01)
            return 'except'
        return 'success'

def main():
    defautl_logger_setup()
    cfg = AioParallelConfig()
    run_aio_parallel_task(cfg, ReqDataBuildProc, DemoAioHttpRunProc)

if __name__ == '__main__':
    main()
